""" Helper methods for EdgePi ADC multiplexer mapping """

import logging

from bitstring import pack
from edgepi.reg_helper.reg_helper import OpCode, BitMask
from edgepi.adc.adc_constants import ADCChannel as CH, AllowedChannels

_logger = logging.getLogger(__name__)

MUXS_PER_ADC = 2
NUM_CHANNELS = 11


class ChannelNotAvailableError(ValueError):
    """
    Raised when an input channel is unavailable for mapping to a multiplexer
    due to RTD_EN status.
    """


def _format_mux_values(mux_p: CH, mux_n: CH):
    # all use cases will always update both mux_n and mux_p
    mask = BitMask.BYTE
    mux_p_val = mux_p.value
    mux_n_val = mux_n.value

    return mux_p_val, mux_n_val, mask


def generate_mux_opcodes(mux_updates: dict):
    """
    Generates list of OpCodes for updating input multiplexer mapping.
    Updates both positive and negative multiplexers.

    Args:
        `mux_updates` (dict): values for updating multiplexer mapping.
            This should be formatted as {ADCReg: (ADCChannel, ADChannel)}.

        `mux_values` (dict): current multiplexer mapping.
            This should be formatted as {ADCReg: (int, int)}.

        Note: both of the above must be dictionaries formatted as:

            mux_reg_addx (ADCReg): (mux_p_val, mux_n_val)

    Returns:
        `list`: OpCodes for updated multiplexer mapping
    """
    _logger.debug(f"mux updates = {mux_updates}")

    mux_opcodes = []
    # generate OpCodes for mux updates
    for addx, byte in mux_updates.items():
        mux_p = byte[0]
        mux_n = byte[1]

        # not updating mux's for this adc_num (no args passed)
        if mux_p is None or mux_n is None:
            continue

        mux_p_val, mux_n_val, mask = _format_mux_values(mux_p, mux_n)

        adc_x_ch_bits = pack("uint:4, uint:4", mux_p_val, mux_n_val).uint

        mux_opcodes.append(OpCode(adc_x_ch_bits, addx.value, mask.value))

    _logger.debug(f"mux opcodes = {mux_opcodes}")

    return mux_opcodes


def validate_channels_allowed(channels: list, rtd_enabled: bool):
    """
    Checks if requested input channels to be mapped are available due to RTD_EN status,
    i.e. in case ADC2 is being used while ADC1 is in RTD mode.

    Args:
        `channels` (list): list of ADCChannel objects representing input channel mapping
        `rtd_enabled` (bool): RTD_EN status
    """
    # channels available depend on RTD_EN status
    allowed_channels = (
        AllowedChannels.RTD_ON.value
        if rtd_enabled
        else AllowedChannels.RTD_OFF.value
    )
    for ch in channels:
        if ch not in allowed_channels:
            raise ChannelNotAvailableError(
                f"Channel 'AIN{ch.value}' is currently not available. Disable RTD in order to use."
            )



"""
Module for GPIO devices
"""


from periphery import GPIO


class GpioDevice:
    """Class for representing a GPIO device"""

    def __init__(self, dev_path: str = None):
        self.fd = dev_path
        self.gpio = None

    def open_gpio(self, pin_num: int = None, pin_dir: str = None, pin_bias: str = None):
        """
        Instantiate GPIO device object for reading and writing.
        Args:
            pin_num (int): pin number to instantiate the object with
            pin_dir (str): pin direction
            pin_bias (str): bias direction
        """
        self.gpio = GPIO(self.fd, pin_num, pin_dir, bias=pin_bias)

    def read_state(self):
        """
        Read the GPIO pin state
        Args:
            N/A
        Return:
            bool: True if high else False
        """
        return self.gpio.read()

    def write_state(self, state: bool = None):
        """
        Write state to GPIO pin
        Args:
            state (bool): High if True else LOW
        Return:
            N/A
        """
        self.gpio.write(state)

    def close(self):
        """Close GPIO connection"""
        self.gpio.close()



"""unit tests for adc_status.py module"""


import pytest
from edgepi.adc.adc_status import get_adc_status, ADCStatusBit, ADCStatusMsg, ADCStatus


@pytest.mark.parametrize(
    "status_byte, expected_status",
    [
        # MSB (bit 7) = index 0 in bitstring
        (
            0b0,
            {
                ADCStatusBit.ADC2_DATA: ADCStatus(
                    ADCStatusBit.ADC2_DATA,
                    ADCStatusMsg.ADC2_DATA_OLD,
                    False
                ),
                ADCStatusBit.ADC1_DATA: ADCStatus(
                    ADCStatusBit.ADC1_DATA,
                    ADCStatusMsg.ADC1_DATA_OLD,
                    False
                ),
                ADCStatusBit.EXTCLK: ADCStatus(
                    ADCStatusBit.EXTCLK,
                    ADCStatusMsg.EXTCLK_INT,
                    False
                ),
                ADCStatusBit.REF_ALM: ADCStatus(
                    ADCStatusBit.REF_ALM,
                    ADCStatusMsg.REF_ALM_OK,
                    False
                ),
                ADCStatusBit.PGAL_ALM: ADCStatus(
                    ADCStatusBit.PGAL_ALM,
                    ADCStatusMsg.PGAL_ALM_OK,
                    False
                ),
                ADCStatusBit.PGAH_ALM: ADCStatus(
                    ADCStatusBit.PGAH_ALM,
                    ADCStatusMsg.PGAH_ALM_OK,
                    False
                ),
                ADCStatusBit.PGAD_ALM: ADCStatus(
                    ADCStatusBit.PGAD_ALM,
                    ADCStatusMsg.PGAD_ALM_OK,
                    False
                ),
                ADCStatusBit.RESET: ADCStatus(
                    ADCStatusBit.RESET,
                    ADCStatusMsg.RESET_FALSE,
                    False
                ),
            },
        ),
        (
            0xFF,
            {
                ADCStatusBit.ADC2_DATA: ADCStatus(
                    ADCStatusBit.ADC2_DATA,
                    ADCStatusMsg.ADC2_DATA_NEW,
                    True
                ),
                ADCStatusBit.ADC1_DATA: ADCStatus(
                    ADCStatusBit.ADC1_DATA,
                    ADCStatusMsg.ADC1_DATA_NEW,
                    True
                ),
                ADCStatusBit.EXTCLK: ADCStatus(
                    ADCStatusBit.EXTCLK,
                    ADCStatusMsg.EXTCLK_EXT,
                    True
                ),
                ADCStatusBit.REF_ALM: ADCStatus(
                    ADCStatusBit.REF_ALM,
                    ADCStatusMsg.REF_ALM_BAD,
                    True
                ),
                ADCStatusBit.PGAL_ALM: ADCStatus(
                    ADCStatusBit.PGAL_ALM,
                    ADCStatusMsg.PGAL_ALM_BAD,
                    True
                ),
                ADCStatusBit.PGAH_ALM: ADCStatus(
                    ADCStatusBit.PGAH_ALM,
                    ADCStatusMsg.PGAH_ALM_BAD,
                    True
                ),
                ADCStatusBit.PGAD_ALM: ADCStatus(
                    ADCStatusBit.PGAD_ALM,
                    ADCStatusMsg.PGAD_ALM_BAD,
                    True
                ),
                ADCStatusBit.RESET: ADCStatus(
                    ADCStatusBit.RESET,
                    ADCStatusMsg.RESET_TRUE,
                    True
                ),
            },
        ),
    ],
)
def test_get_adc_status(status_byte, expected_status):
    assert get_adc_status(status_byte) == expected_status



"""Unit tests for adc_conv_time.py module"""

import pytest
from edgepi.adc.adc_conv_time import (
    expected_initial_time_delay,
    expected_continuous_time_delay,
    ADC1_CONT_DELAYS,
    ADC1_INITIAL_DELAYS,
    ADC2_DELAYS,
)
from edgepi.adc.adc_constants import (
    FilterMode as FILT,
    ADC1DataRate as DR1,
    ADC2DataRate as DR2,
    ADCNum,
)


@pytest.mark.parametrize(
    "adc_num, data_rate, filter_mode, expected",
    [
        (
            ADCNum.ADC_1,
            DR1.SPS_2P5.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2P5.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            # ADC1 PULSE MODE
            ADCNum.ADC_1,
            DR1.SPS_2P5.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2P5.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2P5.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2P5.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2P5.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2P5.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2P5.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2P5.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_5.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_5.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_5.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_5.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_5.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_5.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_5.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_5.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_5.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_5.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_10.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_10.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_10.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_10.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_10.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_10.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_10.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_10.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_10.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_10.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_16P6.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_16P6.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_16P6.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_16P6.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_16P6.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_16P6.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_16P6.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_16P6.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_16P6.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_16P6.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_20.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_20.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_20.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_20.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_20.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_20.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_20.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_20.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_20.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_20.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_50.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_50.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_50.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_50.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_50.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_50.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_50.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_50.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_50.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_50.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_60.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_60.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_60.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_60.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_60.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_60.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_60.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_60.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_60.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_60.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_100.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_100.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_100.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_100.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_100.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_100.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_100.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_100.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_100.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_100.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_400.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_400.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_400.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_400.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_400.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_400.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_400.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_400.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_400.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_400.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_1200.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_1200.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_1200.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_1200.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_1200.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_1200.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_1200.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_1200.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_1200.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_1200.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2400.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2400.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2400.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2400.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2400.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2400.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2400.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2400.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2400.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_2400.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_4800.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_4800.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_4800.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_4800.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_4800.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_4800.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_4800.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_4800.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_4800.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_4800.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_7200.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_7200.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_7200.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_7200.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_7200.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_7200.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_7200.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_7200.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_7200.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_7200.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_14400.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_14400.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_14400.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_14400.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_14400.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_14400.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_14400.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_14400.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_14400.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_14400.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_19200.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_19200.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_19200.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_19200.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_19200.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_19200.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_19200.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_19200.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_19200.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_19200.value.op_code][FILT.FIR.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_38400.value.op_code,
            FILT.SINC1.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_38400.value.op_code][FILT.SINC1.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_38400.value.op_code,
            FILT.SINC2.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_38400.value.op_code][FILT.SINC2.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_38400.value.op_code,
            FILT.SINC3.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_38400.value.op_code][FILT.SINC3.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_38400.value.op_code,
            FILT.SINC4.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_38400.value.op_code][FILT.SINC4.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_38400.value.op_code,
            FILT.FIR.value.op_code,
            ADC1_INITIAL_DELAYS[DR1.SPS_38400.value.op_code][FILT.FIR.value.op_code],
        ),
        # ADC2
        (
            ADCNum.ADC_2,
            DR2.SPS_10.value.op_code,
            FILT.SINC1.value.op_code,
            ADC2_DELAYS[DR2.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_10.value.op_code,
            FILT.SINC2.value.op_code,
            ADC2_DELAYS[DR2.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_10.value.op_code,
            FILT.SINC3.value.op_code,
            ADC2_DELAYS[DR2.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_10.value.op_code,
            FILT.SINC4.value.op_code,
            ADC2_DELAYS[DR2.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_10.value.op_code,
            FILT.FIR.value.op_code,
            ADC2_DELAYS[DR2.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_100.value.op_code,
            FILT.SINC1.value.op_code,
            ADC2_DELAYS[DR2.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_100.value.op_code,
            FILT.SINC2.value.op_code,
            ADC2_DELAYS[DR2.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_100.value.op_code,
            FILT.SINC3.value.op_code,
            ADC2_DELAYS[DR2.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_100.value.op_code,
            FILT.SINC4.value.op_code,
            ADC2_DELAYS[DR2.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_100.value.op_code,
            FILT.FIR.value.op_code,
            ADC2_DELAYS[DR2.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_400.value.op_code,
            FILT.SINC1.value.op_code,
            ADC2_DELAYS[DR2.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_400.value.op_code,
            FILT.SINC2.value.op_code,
            ADC2_DELAYS[DR2.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_400.value.op_code,
            FILT.SINC3.value.op_code,
            ADC2_DELAYS[DR2.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_400.value.op_code,
            FILT.SINC4.value.op_code,
            ADC2_DELAYS[DR2.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_400.value.op_code,
            FILT.FIR.value.op_code,
            ADC2_DELAYS[DR2.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_800.value.op_code,
            FILT.SINC1.value.op_code,
            ADC2_DELAYS[DR2.SPS_800.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_800.value.op_code,
            FILT.SINC2.value.op_code,
            ADC2_DELAYS[DR2.SPS_800.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_800.value.op_code,
            FILT.SINC3.value.op_code,
            ADC2_DELAYS[DR2.SPS_800.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_800.value.op_code,
            FILT.SINC4.value.op_code,
            ADC2_DELAYS[DR2.SPS_800.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_800.value.op_code,
            FILT.FIR.value.op_code,
            ADC2_DELAYS[DR2.SPS_800.value.op_code],
        ),
    ],
)
def test_compute_initial_time_delay(adc_num, data_rate, filter_mode, expected):
    if adc_num == ADCNum.ADC_1:
        assert expected_initial_time_delay(adc_num, data_rate, filter_mode) == expected
    else:
        assert expected_initial_time_delay(adc_num, data_rate, filter_mode) == expected * 3


@pytest.mark.parametrize(
    "adc_num, data_rate, expected",
    [
        # ADC1 CONTINUOUS MODE
        (
            ADCNum.ADC_1,
            DR1.SPS_2P5.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_2P5.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_5.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_5.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_10.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_16P6.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_16P6.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_20.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_20.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_50.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_50.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_60.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_60.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_100.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_400.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_1200.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_1200.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_2400.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_2400.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_4800.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_4800.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_7200.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_7200.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_14400.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_14400.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_19200.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_19200.value.op_code],
        ),
        (
            ADCNum.ADC_1,
            DR1.SPS_38400.value.op_code,
            ADC1_CONT_DELAYS[DR1.SPS_38400.value.op_code],
        ),
        # ADC2
        (
            ADCNum.ADC_2,
            DR2.SPS_10.value.op_code,
            ADC2_DELAYS[DR2.SPS_10.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_100.value.op_code,
            ADC2_DELAYS[DR2.SPS_100.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_400.value.op_code,
            ADC2_DELAYS[DR2.SPS_400.value.op_code],
        ),
        (
            ADCNum.ADC_2,
            DR2.SPS_800.value.op_code,
            ADC2_DELAYS[DR2.SPS_800.value.op_code],
        ),
    ],
)
def test_compute_continuous_time_delay_adc_1(adc_num, data_rate, expected):
    assert expected_continuous_time_delay(adc_num, data_rate) == expected



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eeprom_mapping.proto

# pylint: skip-file
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='eeprom_mapping.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x65\x65prom_mapping.proto\"\xa4\x04\n\x0c\x45\x65promLayout\x12,\n\x03\x64\x61\x63\x18\x01 \x01(\x0b\x32\x1f.EepromLayout.ModuleCalibParams\x12,\n\x03\x61\x64\x63\x18\x02 \x01(\x0b\x32\x1f.EepromLayout.ModuleCalibParams\x12,\n\x03rtd\x18\x03 \x01(\x0b\x32\x1f.EepromLayout.ModuleCalibParams\x12+\n\x02tc\x18\x04 \x01(\x0b\x32\x1f.EepromLayout.ModuleCalibParams\x12(\n\nconfig_key\x18\x05 \x01(\x0b\x32\x14.EepromLayout.AwsKey\x12&\n\x08\x64\x61ta_key\x18\x06 \x01(\x0b\x32\x14.EepromLayout.AwsKey\x12\x15\n\rserial_number\x18\x07 \x01(\t\x12\r\n\x05model\x18\x08 \x01(\t\x12\x11\n\tclient_id\x18\t \x01(\t\x1a\x9d\x01\n\x11ModuleCalibParams\x12<\n\x06\x63\x61libs\x18\x01 \x03(\x0b\x32,.EepromLayout.ModuleCalibParams.ChannelCalib\x1aJ\n\x0c\x43hannelCalib\x12\x11\n\x04gain\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x13\n\x06offset\x18\x02 \x01(\x02H\x01\x88\x01\x01\x42\x07\n\x05_gainB\t\n\x07_offset\x1a\x32\n\x06\x41wsKey\x12\x13\n\x0bprivate_key\x18\x01 \x01(\t\x12\x13\n\x0b\x63\x65rtificate\x18\x02 \x01(\tb\x06proto3'
)




_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB = _descriptor.Descriptor(
  name='ChannelCalib',
  full_name='EepromLayout.ModuleCalibParams.ChannelCalib',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='gain', full_name='EepromLayout.ModuleCalibParams.ChannelCalib.gain', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='offset', full_name='EepromLayout.ModuleCalibParams.ChannelCalib.offset', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_gain', full_name='EepromLayout.ModuleCalibParams.ChannelCalib._gain',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_offset', full_name='EepromLayout.ModuleCalibParams.ChannelCalib._offset',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=447,
  serialized_end=521,
)

_EEPROMLAYOUT_MODULECALIBPARAMS = _descriptor.Descriptor(
  name='ModuleCalibParams',
  full_name='EepromLayout.ModuleCalibParams',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='calibs', full_name='EepromLayout.ModuleCalibParams.calibs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=364,
  serialized_end=521,
)

_EEPROMLAYOUT_AWSKEY = _descriptor.Descriptor(
  name='AwsKey',
  full_name='EepromLayout.AwsKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='private_key', full_name='EepromLayout.AwsKey.private_key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='certificate', full_name='EepromLayout.AwsKey.certificate', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=523,
  serialized_end=573,
)

_EEPROMLAYOUT = _descriptor.Descriptor(
  name='EepromLayout',
  full_name='EepromLayout',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dac', full_name='EepromLayout.dac', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='adc', full_name='EepromLayout.adc', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rtd', full_name='EepromLayout.rtd', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tc', full_name='EepromLayout.tc', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config_key', full_name='EepromLayout.config_key', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_key', full_name='EepromLayout.data_key', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serial_number', full_name='EepromLayout.serial_number', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='EepromLayout.model', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_id', full_name='EepromLayout.client_id', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EEPROMLAYOUT_MODULECALIBPARAMS, _EEPROMLAYOUT_AWSKEY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=573,
)

_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.containing_type = _EEPROMLAYOUT_MODULECALIBPARAMS
_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.oneofs_by_name['_gain'].fields.append(
  _EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.fields_by_name['gain'])
_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.fields_by_name['gain'].containing_oneof = _EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.oneofs_by_name['_gain']
_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.oneofs_by_name['_offset'].fields.append(
  _EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.fields_by_name['offset'])
_EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.fields_by_name['offset'].containing_oneof = _EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB.oneofs_by_name['_offset']
_EEPROMLAYOUT_MODULECALIBPARAMS.fields_by_name['calibs'].message_type = _EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB
_EEPROMLAYOUT_MODULECALIBPARAMS.containing_type = _EEPROMLAYOUT
_EEPROMLAYOUT_AWSKEY.containing_type = _EEPROMLAYOUT
_EEPROMLAYOUT.fields_by_name['dac'].message_type = _EEPROMLAYOUT_MODULECALIBPARAMS
_EEPROMLAYOUT.fields_by_name['adc'].message_type = _EEPROMLAYOUT_MODULECALIBPARAMS
_EEPROMLAYOUT.fields_by_name['rtd'].message_type = _EEPROMLAYOUT_MODULECALIBPARAMS
_EEPROMLAYOUT.fields_by_name['tc'].message_type = _EEPROMLAYOUT_MODULECALIBPARAMS
_EEPROMLAYOUT.fields_by_name['config_key'].message_type = _EEPROMLAYOUT_AWSKEY
_EEPROMLAYOUT.fields_by_name['data_key'].message_type = _EEPROMLAYOUT_AWSKEY
DESCRIPTOR.message_types_by_name['EepromLayout'] = _EEPROMLAYOUT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EepromLayout = _reflection.GeneratedProtocolMessageType('EepromLayout', (_message.Message,), {

  'ModuleCalibParams' : _reflection.GeneratedProtocolMessageType('ModuleCalibParams', (_message.Message,), {

    'ChannelCalib' : _reflection.GeneratedProtocolMessageType('ChannelCalib', (_message.Message,), {
      'DESCRIPTOR' : _EEPROMLAYOUT_MODULECALIBPARAMS_CHANNELCALIB,
      '__module__' : 'eeprom_mapping_pb2'
      # @@protoc_insertion_point(class_scope:EepromLayout.ModuleCalibParams.ChannelCalib)
      })
    ,
    'DESCRIPTOR' : _EEPROMLAYOUT_MODULECALIBPARAMS,
    '__module__' : 'eeprom_mapping_pb2'
    # @@protoc_insertion_point(class_scope:EepromLayout.ModuleCalibParams)
    })
  ,

  'AwsKey' : _reflection.GeneratedProtocolMessageType('AwsKey', (_message.Message,), {
    'DESCRIPTOR' : _EEPROMLAYOUT_AWSKEY,
    '__module__' : 'eeprom_mapping_pb2'
    # @@protoc_insertion_point(class_scope:EepromLayout.AwsKey)
    })
  ,
  'DESCRIPTOR' : _EEPROMLAYOUT,
  '__module__' : 'eeprom_mapping_pb2'
  # @@protoc_insertion_point(class_scope:EepromLayout)
  })
_sym_db.RegisterMessage(EepromLayout)
_sym_db.RegisterMessage(EepromLayout.ModuleCalibParams)
_sym_db.RegisterMessage(EepromLayout.ModuleCalibParams.ChannelCalib)
_sym_db.RegisterMessage(EepromLayout.AwsKey)


# @@protoc_insertion_point(module_scope)



'''
Provides a class for interacting with the GPIO pins through I2C and GPIO peripheral
'''


import logging
from edgepi.gpio.edgepi_gpio_expander import EdgePiGPIOExpander
from edgepi.gpio.edgepi_gpio_chip import EdgePiGPIOChip

_logger = logging.getLogger(__name__)

class PinNameNoneError(Exception):
    """Raised when None is passed"""

class PinNameNotFound(Exception):
    """"Raised when name doesn't exist"""


class EdgePiGPIO(EdgePiGPIOExpander, EdgePiGPIOChip):
    '''
    A class used to represent the GPIO Expander configuration for an I2C Device.
    This class will be imported to each module that requires GPIO manipulation.
    It is not intended for users.
    '''
    def __init__(self):
        _logger.info("GPIO initializing")
        EdgePiGPIOExpander.__init__(self)
        EdgePiGPIOChip.__init__(self)

    def __pin_name_check(self, pin_name: str = None):
        if pin_name is None:
            raise PinNameNoneError(f'Missing Pin name: {pin_name}')
        if pin_name not in self.expander_pin_dict and pin_name not in self.gpiochip_pins_dict:
            raise PinNameNotFound(f'The following pin name: {pin_name} is not found')

    def read_pin_state(self, pin_name: str = None):
        """
        Read corresponding pin state
        Args:
            pin_name (str): name of the pin will be passed as Enum
        return
            state (bool): True/False depending on High/Low
        """
        state = None
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            state = self.read_expander_pin(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            state = self.read_gpio_pin_state(pin_name)
        return state


    def set_pin_state(self, pin_name: str = None):
        """
        Set corresponding pin state to high
        Args:
            pin_name(str): name of the pin will be passed as Enum
        return:
            N/A
        """
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            self.set_expander_pin(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            self.write_gpio_pin_state(pin_name, True)

    def clear_pin_state(self, pin_name: str = None):
        """
        Clearcorresponding pin state to high
        Args:
            pin_name(str): name of the pin will be passed as Enum
        return:
            N/A
        """
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            self.clear_expander_pin(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            self.write_gpio_pin_state(pin_name, False)

    def get_pin_direction(self, pin_name: str = None):
        """
        Get GPIO pin direction
        Args:
            pin_name(str): name of the pin
        Return:
            direction (bool): True if direction is input, False if direction is output
        """
        direction = None
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            direction = self.get_expander_pin_direction(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            direction = bool(self.gpiochip_pins_dict[pin_name].dir == "in")
        return direction

    def set_pin_direction_in(self, pin_name: str = None):
        """
        Set GPIO pin direction to input
        Args:
            pin_name(str): name of the pin
        Return:
            N/A
        """
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            self.set_expander_pin_direction_in(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            self.set_gpio_pin_dir(pin_name, True)

    def set_pin_direction_out(self, pin_name: str = None):
        """
        Set GPIO pin direction to output
        Args:
            pin_name(str): name of the pin
        Return:
            N/A
        """
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            self.set_expander_pin_direction_out(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            self.set_gpio_pin_dir(pin_name, False)

    def toggle_pin(self, pin_name: str = None):
        """
        Toggle GPIO pin
        Args:
            pin_name (str): name of the pin
        Return:
            N/A
        """
        self.__pin_name_check(pin_name)
        if pin_name in self.expander_pin_dict:
            self.toggle_expander_pin(pin_name)
        if pin_name in self.gpiochip_pins_dict:
            self.toggle_gpio_pin_state(pin_name)



'''Address map of eeprom'''

from enum import Enum
from dataclasses import dataclass
from edgepi.calibration.eeprom_mapping_pb2 import EepromLayout
from edgepi.calibration.calibration_constants import CalibParam

class EEPROMInfo(Enum):
    """
    EEPROM device address for I2C addressing
    """
    DEV_ADDR = 0x50
    PAGE_SIZE = 64
    NUM_OF_PAGE = 512

class ModuleNames(Enum):
    """
    Module Name Enum
    """
    DAC = 0x0
    ADC = 0x1
    RTD = 0x2
    TC = 0x3

class EdgePiMemoryInfo(Enum):
    """
    Information regarding Edgepi reserved memory space
    """
    USED_SPACE = 0x00
    BUFF_START = 0x02

class MessageFieldNumber(Enum):
    """
    MessageField index number to be used for ListFields() function. The function lists fields start
    from index 0
    """
    DAC=1
    ADC=2
    RTD=3
    TC=4
    CONFIGS_KEY=5
    DATA_KEY=6
    SERIAL=7
    MODEL=8
    CLIENT_ID=9

@dataclass
class Keys:
    """
    Dataclass to store key strings
    """
    certificate: str = None
    private: str = None

@dataclass
class EdgePiEEPROMData:
    # pylint: disable=too-many-instance-attributes
    """
    Dataclass to store edgepi reserved values
    dac_calib_parms (dict): list of calibration parameters
    adc_calib_parms (dict): list of calibration parameters
    rtd_calib_parms (dict): list of calibration parameters
    tc_calib_parms (dict): list of calibration parameters
    config_key (Keys): dataclass
    data_key (Keys): dataclass
    serial (str)
    model (str)
    client_id (str)
    """
    dac_calib_parms: dict = None
    adc_calib_parms: dict = None
    rtd_calib_parms: dict = None
    tc_calib_parms: dict = None
    config_key: Keys = None
    data_key: Keys = None
    serial: str = None
    model: str = None
    client_id: str = None

    def message_to_dict(self, data_to_unpack: EepromLayout = None):
        """
        Function to unpack message to list
        Args:
            data_to_unpack: EepromLayout message modules
        Returns:
            calib_list: 1-D array
        """
        calib_dict={}
        for indx, ch in enumerate(data_to_unpack.calibs):
            calib_dict[indx] = CalibParam(gain=ch.gain,
                                        offset=ch.offset)
        return calib_dict

    def keys_to_str(self, data_to_unpack: EepromLayout = None):
        """
        Function to unpack message to string
        Args:
            data_to_unpack: EepromLayout message keys
        Returns:
            Keys (dataclass): keys values
        """
        return Keys(certificate = data_to_unpack.certificate, private = data_to_unpack.private_key)



"""Utility module for ADC fault reading"""

from dataclasses import dataclass
from enum import Enum, unique

import bitstring


@unique
class ADCStatusBit(Enum):
    """
    Bit numbers for each ADC STATUS byte fault/field (in reverse order)
    """

    RESET = 7
    PGAD_ALM = 6
    PGAH_ALM = 5
    PGAL_ALM = 4
    REF_ALM = 3
    EXTCLK = 2
    ADC1_DATA = 1
    ADC2_DATA = 0


@unique
class ADCStatusMsg(Enum):
    """Debugging messages used in STATUS byte info returned to the user"""

    ADC2_DATA_NEW = "New voltage read data is available for ADC2"
    ADC2_DATA_OLD = "No new voltage read data is available for ADC2"
    ADC1_DATA_NEW = "New voltage read data is available for ADC1"
    ADC1_DATA_OLD = "No new voltage read data is available for ADC1"
    EXTCLK_EXT = "ADC clock source is external"
    EXTCLK_INT = "ADC clock source is internal"
    REF_ALM_OK = "ADC1: no low reference voltage alarm detected"
    REF_ALM_BAD = "ADC1: low reference voltage alarm detected, less than or equal to 0.4 V"
    PGAL_ALM_OK = "ADC1: no absolute low voltage alarm detected"
    PGAL_ALM_BAD = "ADC1: absolute low voltage alarm detected"
    PGAH_ALM_OK = "ADC1: no absolute high voltage alarm detected"
    PGAH_ALM_BAD = "ADC1: absolute high voltage alarm detected"
    PGAD_ALM_OK = "ADC1: no PGA differential output range alarm detected"
    PGAD_ALM_BAD = "ADC1: PGA differential output range alarm detected"
    RESET_TRUE = "New ADC device reset has occured"
    RESET_FALSE = "No new ADC device reset has occured since RESET bit last cleared"



@dataclass
class ADCStatus:
    """Represents the status of a STATUS byte field

    Attributes:
        name (str): the STATUS byte field's official ADS1263 name

        status_type (ADCStatusBit): the ADCFaultType STATUS bit type

        err_msg (ADCStatusMsg): a message containing the current fault output

        at_fault (bool): set to True if the STATUS bit is set
    """

    status_type: ADCStatusBit = None
    err_msg: ADCStatusMsg = None
    at_fault: bool = False

    def __repr__(self) -> str:
        msg = (
            "\n\t{"
            + f"""
            STATUS bit: {self.status_type},
            At Fault: {self.at_fault},
            Status Message: {self.err_msg.value},
        """
            + "}\n"
        )
        return msg


_fault_msg_map = {
    ADCStatusBit.ADC2_DATA: (ADCStatusMsg.ADC2_DATA_OLD, ADCStatusMsg.ADC2_DATA_NEW),
    ADCStatusBit.ADC1_DATA: (ADCStatusMsg.ADC1_DATA_OLD, ADCStatusMsg.ADC1_DATA_NEW),
    ADCStatusBit.EXTCLK: (ADCStatusMsg.EXTCLK_INT, ADCStatusMsg.EXTCLK_EXT),
    ADCStatusBit.REF_ALM: (ADCStatusMsg.REF_ALM_OK, ADCStatusMsg.REF_ALM_BAD),
    ADCStatusBit.PGAL_ALM: (ADCStatusMsg.PGAL_ALM_OK, ADCStatusMsg.PGAL_ALM_BAD),
    ADCStatusBit.PGAH_ALM: (ADCStatusMsg.PGAH_ALM_OK, ADCStatusMsg.PGAH_ALM_BAD),
    ADCStatusBit.PGAD_ALM: (ADCStatusMsg.PGAD_ALM_OK, ADCStatusMsg.PGAD_ALM_BAD),
    ADCStatusBit.RESET: (ADCStatusMsg.RESET_FALSE, ADCStatusMsg.RESET_TRUE),
}


def get_adc_status(status_code: int) -> dict:
    """Generates a dictionary of ADC Status objects

    Args:
        `status_byte` (int): uint value of ADC STATUS byte from voltage reading

    Returns:
        `dict`: contains information on the current status of each STATUS byte bit
    """
    status_dict = {}

    status_byte = bitstring.pack("uint:8", status_code)

    # check each bit in status_byte
    for bit_num in ADCStatusBit:
        status = ADCStatus(status_type=bit_num)

        # get value of ith bit in status byte, either 0 or 1
        bit_value = status_byte[bit_num.value]

        # get message depending on whether ith bit of status byte is set or not
        status.err_msg = _fault_msg_map[bit_num][bit_value]

        status.at_fault = bit_value

        status_dict[bit_num] = status

    return status_dict



""" Unit tests for edgepi_dac module """

from copy import deepcopy
from unittest import mock
from unittest.mock import call
import sys

sys.modules['periphery'] = mock.MagicMock()

# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import pytest
from bitstring import pack
from edgepi.dac.dac_constants import (
    AOPins,
    PowerMode,
    DACChannel as CH,
    EdgePiDacCom as COM,
)
from edgepi.dac.edgepi_dac import EdgePiDAC
from edgepi.calibration.eeprom_constants import EdgePiEEPROMData
from edgepi.calibration.calibration_constants import CalibParam

dummy_calib_param_dict = {0:CalibParam(gain=1,offset=0),
                          1:CalibParam(gain=1,offset=0),
                          2:CalibParam(gain=1,offset=0),
                          3:CalibParam(gain=1,offset=0),
                          4:CalibParam(gain=1,offset=0),
                          5:CalibParam(gain=1,offset=0),
                          6:CalibParam(gain=1,offset=0),
                          7:CalibParam(gain=1,offset=0)}

@pytest.fixture(name="dac")
def fixture_test_dac(mocker):
    mocker.patch("edgepi.peripherals.spi.SPI")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiEEPROM.get_edgepi_reserved_data",
                  return_value = EdgePiEEPROMData(dac_calib_parms=dummy_calib_param_dict))
    yield EdgePiDAC()

@pytest.fixture(name="dac_mock_periph")
def fixture_test_dac_write_voltage(mocker):
    mocker.patch("edgepi.peripherals.spi.SPI")
    mocker.patch("edgepi.peripherals.i2c.I2C")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiEEPROM.get_edgepi_reserved_data",
                  return_value = EdgePiEEPROMData(dac_calib_parms=dummy_calib_param_dict))
    yield EdgePiDAC()


_default_power_modes = {
    CH.AOUT7.value: PowerMode.NORMAL.value,
    CH.AOUT6.value: PowerMode.NORMAL.value,
    CH.AOUT5.value: PowerMode.NORMAL.value,
    CH.AOUT4.value: PowerMode.NORMAL.value,
    CH.AOUT3.value: PowerMode.NORMAL.value,
    CH.AOUT2.value: PowerMode.NORMAL.value,
    CH.AOUT1.value: PowerMode.NORMAL.value,
    CH.AOUT0.value: PowerMode.NORMAL.value,
}


@pytest.mark.parametrize(
    "power_modes, analog_out, mode, expected",
    [
        (deepcopy(_default_power_modes), CH.AOUT7, PowerMode.POWER_DOWN_GROUND, [64, 64, 0]),
        (deepcopy(_default_power_modes), CH.AOUT7, PowerMode.POWER_DOWN_3_STATE, [64, 192, 0]),
        (deepcopy(_default_power_modes), CH.AOUT7, PowerMode.NORMAL, [64, 0, 0]),
        (
            {
                CH.AOUT7.value: PowerMode.POWER_DOWN_GROUND.value,
                CH.AOUT6.value: PowerMode.NORMAL.value,
                CH.AOUT5.value: PowerMode.NORMAL.value,
                CH.AOUT4.value: PowerMode.NORMAL.value,
                CH.AOUT3.value: PowerMode.NORMAL.value,
                CH.AOUT2.value: PowerMode.NORMAL.value,
                CH.AOUT1.value: PowerMode.NORMAL.value,
                CH.AOUT0.value: PowerMode.NORMAL.value,
            },
            CH.AOUT7,
            PowerMode.NORMAL,
            [64, 0, 0],
        ),
        (
            {
                CH.AOUT7.value: PowerMode.NORMAL.value,
                CH.AOUT6.value: PowerMode.NORMAL.value,
                CH.AOUT5.value: PowerMode.POWER_DOWN_GROUND.value,
                CH.AOUT4.value: PowerMode.NORMAL.value,
                CH.AOUT3.value: PowerMode.NORMAL.value,
                CH.AOUT2.value: PowerMode.NORMAL.value,
                CH.AOUT1.value: PowerMode.NORMAL.value,
                CH.AOUT0.value: PowerMode.NORMAL.value,
            },
            CH.AOUT5,
            PowerMode.NORMAL,
            [0x40, 0, 0],
        ),
    ],
)
def test_dac_set_power_mode(mocker, power_modes, analog_out, mode, expected, dac):
    mocker.patch.object(dac, "_EdgePiDAC__dac_power_state", power_modes)
    mock_transfer = mocker.patch("edgepi.peripherals.spi.SpiDevice.transfer")
    dac.set_power_mode(analog_out, mode)
    mock_transfer.assert_called_once_with(expected)


def test_dac_reset(mocker, dac):
    mock_transfer = mocker.patch("edgepi.peripherals.spi.SpiDevice.transfer")
    dac.reset()
    mock_transfer.assert_called_once_with([96, 18, 52])

@pytest.mark.parametrize("analog_out, mock_val",
                         [(CH.AOUT0, [0xA1,0x69, 0xDF]),
                          (CH.AOUT1, [0xA1,0x69, 0xDF]),
                          (CH.AOUT2, [0xA1,0x69, 0xDF]),
                          (CH.AOUT3, [0xA1,0x69, 0xDF]),
                          (CH.AOUT4, [0xA1,0x69, 0xDF]),
                          (CH.AOUT5, [0xA1,0x69, 0xDF]),
                          (CH.AOUT6, [0xA1,0x69, 0xDF]),
                          (CH.AOUT7, [0xA1,0x69, 0xDF])])
def test_channel_readback(mocker, analog_out, mock_val, dac):
    mocker.patch("edgepi.peripherals.spi.SpiDevice.transfer", return_value=mock_val)
    bits = pack("uint:8, uint:8, uint:8", mock_val[0], mock_val[1], mock_val[2])
    result = bits[-16:].uint
    code = dac.channel_readback(analog_out)
    assert code == result

@pytest.mark.parametrize(
    "analog_out, read_data",
    [
        (CH.AOUT7, [0, 0, 0]),
        (CH.AOUT6, [0, 0, 0]),
        (CH.AOUT5, [0, 0, 0]),
        (CH.AOUT4, [0, 0, 0]),
        (CH.AOUT3, [0, 0, 0]),
        (CH.AOUT2, [0, 0, 0]),
        (CH.AOUT1, [0, 0, 0]),
        (CH.AOUT0, [0, 0, 0]),
    ]
)
def test_dac_compute_expected_voltage(mocker, analog_out, read_data, dac):

    mock_transfer = mocker.patch(
        "edgepi.peripherals.spi.SpiDevice.transfer", return_value=read_data
    )
    byte_1 = (COM.COM_READBACK.value << 4) + analog_out.value
    dac.compute_expected_voltage(analog_out)
    write_calls = [call([byte_1, 0, 0]), call([0, 0, 0])]
    mock_transfer.assert_has_calls(write_calls)


@pytest.mark.parametrize(
    "analog_out, pin_name, voltage, mock_name",
    [
        (0, AOPins.AO_EN1.value, 1.0, "mock_set"),
        (1, AOPins.AO_EN2.value, 1.0, "mock_set"),
        (2, AOPins.AO_EN3.value, 1.0, "mock_set"),
        (3, AOPins.AO_EN4.value, 1.0, "mock_set"),
        (4, AOPins.AO_EN5.value, 1.0, "mock_set"),
        (5, AOPins.AO_EN6.value, 1.0, "mock_set"),
        (6, AOPins.AO_EN7.value, 1.0, "mock_set"),
        (7, AOPins.AO_EN8.value, 1.0, "mock_set"),
        (0, AOPins.AO_EN1.value, 0, "mock_clear"),
        (1, AOPins.AO_EN2.value, 0, "mock_clear"),
        (2, AOPins.AO_EN3.value, 0, "mock_clear"),
        (3, AOPins.AO_EN4.value, 0, "mock_clear"),
        (4, AOPins.AO_EN5.value, 0, "mock_clear"),
        (5, AOPins.AO_EN6.value, 0, "mock_clear"),
        (6, AOPins.AO_EN7.value, 0, "mock_clear"),
        (7, AOPins.AO_EN8.value, 0, "mock_clear"),
    ]
)
def test_dac_send_to_gpio_pins(mocker, analog_out, pin_name, voltage, mock_name):
    # can't mock entire GPIO class here because need to access its methods
    mocker.patch("edgepi.peripherals.spi.SPI")
    mocker.patch("edgepi.peripherals.i2c.I2C")
    mocker.patch("edgepi.gpio.edgepi_gpio_expander.I2CDevice")
    mock_set = mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.set_expander_pin")
    mock_clear = mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.clear_expander_pin")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiEEPROM.get_edgepi_reserved_data",
                  return_value = EdgePiEEPROMData(dac_calib_parms=dummy_calib_param_dict))
    dac = EdgePiDAC()
    dac._EdgePiDAC__send_to_gpio_pins(analog_out, voltage)
    # check correct clause is entered depending on voltage written
    if voltage > 0:
        mock_set.assert_called_with(pin_name)
        mock_set.name = "mock_set"
        assert mock_set.name == mock_name
    else:
        mock_clear.assert_called_with(pin_name)
        mock_clear.name = "mock_clear"
        assert mock_clear.name == mock_name


@pytest.mark.parametrize('analog_out, voltage', [
    (0, -0.1),
    (1, -0.1),
    (2, -0.1),
    (3, -0.1),
    (4, -0.1),
    (5, -0.1),
    (6, -0.1),
    (7, -0.1),
])
def test_send_to_gpio_pins_raises(analog_out, voltage, dac):
    with pytest.raises(ValueError) as err:
        dac._EdgePiDAC__send_to_gpio_pins(analog_out, voltage)
        assert err.value == "voltage cannot be negative"


@pytest.mark.parametrize("analog_out, voltage, mock_value, result",
                         [(CH.AOUT0, 2.123, [None, None, True], [13913]),
                          (CH.AOUT1, 2.123, [None, None, True], [13913]),
                          (CH.AOUT2, 2.123, [None, None, True], [13913]),
                          (CH.AOUT3, 2.123, [None, None, True], [13913]),
                          (CH.AOUT4, 2.123, [None, None, False], [27826]),
                          (CH.AOUT5, 2.123, [None, None, False], [27826]),
                          (CH.AOUT6, 2.123, [None, None, False], [27826]),
                          (CH.AOUT7, 2.123, [None, None, False], [27826])
                        ])
def test_write_voltage(mocker,analog_out, voltage, mock_value, result, dac_mock_periph):
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiDAC.get_state",
                  return_value = (mock_value[0], mock_value[1], mock_value[2]))
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.get_pin_direction", return_value = False)
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.set_expander_pin")
    assert result[0] == dac_mock_periph.write_voltage(analog_out, voltage)

@pytest.mark.parametrize("enable, result, mocker_values,",
                        [(True, True, [None, None, True]),
                         (False, False, [None, None, False])])
def test_enable_dac_gain(mocker, enable, result, mocker_values):
    mocker.patch("edgepi.peripherals.spi.SPI")
    mocker.patch("edgepi.peripherals.i2c.I2C")
    mocker.patch("edgepi.gpio.edgepi_gpio_expander.I2CDevice")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiDAC.get_state",
                  return_value = (mocker_values[0], mocker_values[1], mocker_values[2]))
    set_dac_gain = mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.set_expander_pin")
    clear_dac_gain = mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.clear_expander_pin")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiEEPROM.get_edgepi_reserved_data",
                  return_value = EdgePiEEPROMData(dac_calib_parms=dummy_calib_param_dict))
    dac = EdgePiDAC()
    # pylint: disable=expression-not-assigned
    assert dac.enable_dac_gain(enable) == result
    set_dac_gain.assert_called_once_with("DAC_GAIN") if enable \
        else clear_dac_gain.assert_called_once_with("DAC_GAIN")

@pytest.mark.parametrize("mock_val, analog_out, code, voltage, gain, result",
    [([0xA1,0x69, 0xDF, True], CH.AOUT0, True, True, True, [27103, 2.116, True]),
    ([0xA1,0x69, 0xDF, False], CH.AOUT1, True, True, True, [27103, 2.116, False]),
    ([0xA1,0x69, 0xDF, True], CH.AOUT2, True, True, False, [27103, 2.116, None]),
    ([0xA1,0x69, 0xDF, True], CH.AOUT3, True, True, None, [27103, 2.116, None]),
    ([0xA1,0x69, 0xDF, True], CH.AOUT4, True, False, True, [27103, None, True]),
    ([0xA1,0x69, 0xDF, True], CH.AOUT5, True, None, True, [27103, None, True]),
    ([0xA1,0x69, 0xDF, True], CH.AOUT6, False, True, True, [None, 2.116, True]),
    ([0xA1,0x69, 0xDF, True], CH.AOUT7, None, True, True, [None, 2.116, True])])
def test_get_state(mocker, analog_out, code, voltage, gain, result, mock_val):
    mocker.patch("edgepi.peripherals.spi.SPI")
    mocker.patch("edgepi.peripherals.i2c.I2C")
    mocker.patch("edgepi.gpio.edgepi_gpio_expander.I2CDevice")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiGPIO.get_pin_direction", return_value = mock_val[3])
    mocker.patch("edgepi.peripherals.spi.SpiDevice.transfer", return_value=mock_val[0:3])
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiEEPROM.get_edgepi_reserved_data",
                  return_value = EdgePiEEPROMData(dac_calib_parms=dummy_calib_param_dict))
    dac = EdgePiDAC()
    code_val, voltage_val, gain_state = dac.get_state(analog_out, code, voltage, gain)
    assert code_val == result[0]
    assert pytest.approx(voltage_val, 1e-3) == voltage_val
    assert gain_state == result[2]



"Unit testing ADCState class"

from unittest import mock
from copy import deepcopy
import sys
import pytest

sys.modules['periphery'] = mock.MagicMock()

# pylint: disable=wrong-import-position
# mocked periphery module needs to be placed above
from edgepi.adc.edgepi_adc import ADCState
from edgepi.adc.adc_query_lang import ADCProperties
from edgepi.adc.adc_constants import (
    ADC1DataRate,
    ADC2DataRate,
    ADCChannel,
    ADCReg,
    CheckMode,
    ConvMode,
    FilterMode,
    StatusByte,
)

# mock default register values
ADC_REGS = {
    ADCReg.REG_ID.value: 0x0,
    ADCReg.REG_POWER.value: 0x0,
    ADCReg.REG_INTERFACE.value: 0x0,
    ADCReg.REG_MODE0.value: 0x0,
    ADCReg.REG_MODE1.value: 0x0,
    ADCReg.REG_MODE2.value: 0x0,
    ADCReg.REG_INPMUX.value: 0x0,
    ADCReg.REG_OFCAL0.value: 0x0,
    ADCReg.REG_OFCAL1.value: 0x0,
    ADCReg.REG_OFCAL2.value: 0x0,
    ADCReg.REG_FSCAL0.value: 0x0,
    ADCReg.REG_FSCAL1.value: 0x0,
    ADCReg.REG_FSCAL2.value: 0x0,
    ADCReg.REG_IDACMUX.value: 0x0,
    ADCReg.REG_IDACMAG.value: 0x0,
    ADCReg.REG_REFMUX.value: 0x0,
    ADCReg.REG_TDACP.value: 0x0,
    ADCReg.REG_TDACN.value: 0x0,
    ADCReg.REG_GPIOCON.value: 0x0,
    ADCReg.REG_GPIODIR.value: 0x0,
    ADCReg.REG_GPIODAT.value: 0x0,
    ADCReg.REG_ADC2CFG.value: 0x0,
    ADCReg.REG_ADC2MUX.value: 0x0,
    ADCReg.REG_ADC2OFC0.value: 0x0,
    ADCReg.REG_ADC2OFC1.value: 0x0,
    ADCReg.REG_ADC2FSC0.value: 0x0,
    ADCReg.REG_ADC2FSC1.value: 0x0,
}


def _apply_register_updates(reg_map: dict, updates: dict):
    for addx, value in updates.items():
        reg_map[addx] = value


@pytest.mark.parametrize(
    "updates, state_property, expected",
    [
        # CHECK_MODE
        (
            {ADCReg.REG_INTERFACE.value: 0x2},
            "state.checksum_mode",
            ADCProperties.CHECK_MODE.value.values[CheckMode.CHECK_BYTE_CRC.value.op_code],
        ),
        (
            {ADCReg.REG_INTERFACE.value: 0x1},
            "state.checksum_mode",
            ADCProperties.CHECK_MODE.value.values[CheckMode.CHECK_BYTE_CHK.value.op_code],
        ),
        (
            {ADCReg.REG_INTERFACE.value: 0x0},
            "state.checksum_mode",
            ADCProperties.CHECK_MODE.value.values[CheckMode.CHECK_BYTE_OFF.value.op_code],
        ),
        # CONV_MODE
        (
            {ADCReg.REG_MODE0.value: 0x40},
            "state.adc_1.conversion_mode",
            ADCProperties.CONV_MODE.value.values[ConvMode.PULSE.value.op_code],
        ),
        (
            {ADCReg.REG_MODE0.value: 0x00},
            "state.adc_1.conversion_mode",
            ADCProperties.CONV_MODE.value.values[ConvMode.CONTINUOUS.value.op_code],
        ),
        # STATUS_MODE
        (
            {ADCReg.REG_INTERFACE.value: 0x0},
            "state.status_byte",
            ADCProperties.STATUS_MODE.value.values[StatusByte.STATUS_BYTE_OFF.value.op_code],
        ),
        (
            {ADCReg.REG_INTERFACE.value: 0x4},
            "state.status_byte",
            ADCProperties.STATUS_MODE.value.values[StatusByte.STATUS_BYTE_ON.value.op_code],
        ),
        # ADC1_MUXP
        (
            {ADCReg.REG_INPMUX.value: 0x00},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN0.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x10},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN1.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x20},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN2.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x30},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN3.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x40},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN4.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x50},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN5.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x60},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN6.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x70},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN7.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x80},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN8.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x90},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AIN9.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0xA0},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.AINCOM.value << 4],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0xF0},
            "state.adc_1.mux_p",
            ADCProperties.ADC1_MUXP.value.values[ADCChannel.FLOAT.value << 4],
        ),
        # ADC1_MUXN
        (
            {ADCReg.REG_INPMUX.value: 0x00},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN0.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x1},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN1.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x2},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN2.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x3},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN3.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x4},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN4.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x5},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN5.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x6},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN6.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x7},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN7.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x8},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN8.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0x9},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AIN9.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0xA},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.AINCOM.value],
        ),
        (
            {ADCReg.REG_INPMUX.value: 0xF},
            "state.adc_1.mux_n",
            ADCProperties.ADC1_MUXN.value.values[ADCChannel.FLOAT.value],
        ),
        # ADC2_MUXP
        (
            {ADCReg.REG_ADC2MUX.value: 0x00},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN0.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x10},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN1.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x20},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN2.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x30},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN3.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x40},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN4.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x50},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN5.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x60},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN6.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x70},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN7.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x80},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN8.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x90},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AIN9.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0xA0},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.AINCOM.value << 4],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0xF0},
            "state.adc_2.mux_p",
            ADCProperties.ADC2_MUXP.value.values[ADCChannel.FLOAT.value << 4],
        ),
        # ADC2_MUXN
        (
            {ADCReg.REG_ADC2MUX.value: 0x00},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN0.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x1},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN1.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x2},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN2.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x3},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN3.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x4},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN4.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x5},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN5.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x6},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN6.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x7},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN7.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x8},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN8.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0x9},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AIN9.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0xA},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.AINCOM.value],
        ),
        (
            {ADCReg.REG_ADC2MUX.value: 0xF},
            "state.adc_2.mux_n",
            ADCProperties.ADC2_MUXN.value.values[ADCChannel.FLOAT.value],
        ),
        # FILTER_MODE
        (
            {ADCReg.REG_MODE1.value: 0x0},
            "state.filter_mode",
            ADCProperties.FILTER_MODE.value.values[FilterMode.SINC1.value.op_code],
        ),
        (
            {ADCReg.REG_MODE1.value: 0b00100000},
            "state.filter_mode",
            ADCProperties.FILTER_MODE.value.values[FilterMode.SINC2.value.op_code],
        ),
        (
            {ADCReg.REG_MODE1.value: 0b01000000},
            "state.filter_mode",
            ADCProperties.FILTER_MODE.value.values[FilterMode.SINC3.value.op_code],
        ),
        (
            {ADCReg.REG_MODE1.value: 0b01100000},
            "state.filter_mode",
            ADCProperties.FILTER_MODE.value.values[FilterMode.SINC4.value.op_code],
        ),
        (
            {ADCReg.REG_MODE1.value: 0b10000000},
            "state.filter_mode",
            ADCProperties.FILTER_MODE.value.values[FilterMode.FIR.value.op_code],
        ),
        # RTD on
        (
            {
                ADCReg.REG_INPMUX.value: 0x56,
                ADCReg.REG_IDACMUX.value: 0x98,
                ADCReg.REG_IDACMAG.value: 0x44,
                ADCReg.REG_REFMUX.value: 0b00011100,
            },
            "state.rtd_on",
            True
        ),
        # RTD off
        (
            {
                ADCReg.REG_INPMUX.value: 0xFA,
                ADCReg.REG_IDACMUX.value: 0xBB,
                ADCReg.REG_IDACMAG.value: 0x00,
                ADCReg.REG_REFMUX.value: 0x00,
            },
            "state.rtd_on",
            False
        ),
        # RTD improperly configured
        (
            {
                ADCReg.REG_INPMUX.value: 0x56,
                ADCReg.REG_IDACMUX.value: 0x98,
                ADCReg.REG_IDACMAG.value: 0x48,
                ADCReg.REG_REFMUX.value: 0b00011100,
            },
            "state.rtd_on",
            False
        ),
        # ADC1 Data Rate
        (
            {ADCReg.REG_MODE2.value: 0x00},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_2P5.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x01},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_5.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x02},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_10.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x03},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_16P6.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x04},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_20.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x05},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_50.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x06},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_60.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x07},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_100.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x08},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_400.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x09},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_1200.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x0A},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_2400.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x0B},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_4800.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x0C},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_7200.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x0D},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_14400.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x0E},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_19200.value.op_code],
        ),
        (
            {ADCReg.REG_MODE2.value: 0x0F},
            "state.adc_1.data_rate",
            ADCProperties.DATA_RATE_1.value.values[ADC1DataRate.SPS_38400.value.op_code],
        ),
        # ADC2 Data Rate
        (
            {ADCReg.REG_ADC2CFG.value: 0x00},
            "state.adc_2.data_rate",
            ADCProperties.DATA_RATE_2.value.values[ADC2DataRate.SPS_10.value.op_code],
        ),
        (
            {ADCReg.REG_ADC2CFG.value: 0x40},
            "state.adc_2.data_rate",
            ADCProperties.DATA_RATE_2.value.values[ADC2DataRate.SPS_100.value.op_code],
        ),
        (
            {ADCReg.REG_ADC2CFG.value: 0x80},
            "state.adc_2.data_rate",
            ADCProperties.DATA_RATE_2.value.values[ADC2DataRate.SPS_400.value.op_code],
        ),
        (
            {ADCReg.REG_ADC2CFG.value: 0xC0},
            "state.adc_2.data_rate",
            ADCProperties.DATA_RATE_2.value.values[ADC2DataRate.SPS_800.value.op_code],
        ),
    ],
)
def test_adc_state_init(updates, state_property, expected):
    reg_map = deepcopy(ADC_REGS)
    _apply_register_updates(reg_map, updates)
    # pylint: disable=eval-used, unused-variable
    # using eval to access nested attributes of state with dot notation
    state = ADCState(reg_map)
    assert eval(state_property) == expected



"""unit tests for edgepi_gpio_chip module"""

# pylint: disable=protected-access
# pylint: disable=C0413

from unittest import mock
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()

import pytest
from edgepi.gpio.gpio_configs import generate_gpiochip_pin_info
from edgepi.gpio.edgepi_gpio_chip import EdgePiGPIOChip


@pytest.fixture(name='mock_gpio')
def fixture_mock_i2c_lib(mocker):
    yield mocker.patch('edgepi.peripherals.gpio.GPIO')

def test_edgepi_gpio_init():
    gpio = EdgePiGPIOChip()
    assert gpio.gpiochip_pins_dict == generate_gpiochip_pin_info()

@pytest.mark.parametrize("pin_name, mock_value, result", [("DIN1",[True], True)])
def test_read_gpio_pin_state(mocker, pin_name, mock_value, result):
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.open_gpio")
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.close")
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.read_state", return_value = mock_value[0])
    gpio = EdgePiGPIOChip()
    assert gpio.read_gpio_pin_state(pin_name) == result

@pytest.mark.parametrize("pin_name, mock_value, result", [("DIN1",[True], True),
                                                          ("DIN1",[False], False)])
def test_write_gpio_pin_state(mocker, pin_name, mock_value, result):
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.open_gpio")
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.close")
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.write_state")
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.read_state", return_value = mock_value[0])
    gpio = EdgePiGPIOChip()
    assert gpio.write_gpio_pin_state(pin_name, mock_value[0]) == result



"""'
    Accessory module for EdgePiTC methods

    Classes:


    Functions:
        _negative_temp_check(int)
        code_to_temp(list)
"""


import logging
from enum import Enum, unique
from dataclasses import dataclass
from bitstring import BitArray, pack
from edgepi.tc.tc_constants import REG_SIZE, TCAddresses, TCType, TempBits, Masks
from edgepi.reg_helper.reg_helper import OpCode

_logger = logging.getLogger(__name__)


def _negative_temp_check(temp_code: int):
    """checks if MAX31856 temperature register reading is negative"""
    # if sign bit is set (negative temp), toggle sign bit off. Compute as
    # positive and then invert after.
    if temp_code[0]:
        del temp_code[0]
        return True
    return False


def code_to_temp(code_bytes: list):
    """
    converts cold junction and linearized thermocouple temperature
    binary codes to float values
    """
    # generate cold junction temperature and linearized TC temperatures
    cj_high_byte = code_bytes[1]
    cj_low_byte = code_bytes[2]
    lt_high_byte = code_bytes[3]
    lt_mid_byte = code_bytes[4]
    lt_low_byte = code_bytes[5]

    cj_negative_flag, lt_negative_flag = False, False

    # combine and arrange the temperature bits in order from msb to lsb and
    # eliminate empty placeholder bits
    cj_code = pack("uint:8, uint:8", cj_high_byte, cj_low_byte)
    del cj_code[-2:]
    cj_negative_flag = _negative_temp_check(cj_code)

    # combine and arrange the temperature bits in order from msb to lsb and
    # eliminate empty placeholder bits
    lt_code = pack("uint:8, uint:8, uint:8", lt_high_byte, lt_mid_byte, lt_low_byte)
    del lt_code[-5:]
    lt_negative_flag = _negative_temp_check(lt_code)

    # >> shift the 6 precision bits behind the decimal point
    cj_temp = cj_code.uint * (2**-TempBits.CJ_DECIMAL_BITS.value)
    # >> shift the 7 precision bits behind the decimal point
    lt_temp = lt_code.uint * (2**-TempBits.LT_DECIMAL_BITS.value)

    # invert if negative temperature
    if cj_negative_flag:
        cj_temp = -(cj_temp)
    if lt_negative_flag:
        lt_temp = -(lt_temp)

    _logger.info(f"Cold Junction Temp: {cj_temp}")
    _logger.info(f"THERMOCOUPLE TC Temp: {lt_temp}")

    return cj_temp, lt_temp


@unique
class TempType(Enum):
    """MAX31856 temperature threshold setting types"""

    COLD_JUNCTION = 0
    THERMOCOUPLE = 1
    COLD_JUNCTION_OFFSET = 2


@dataclass
class TempCode:
    """Holds data needed to update a MAX31856 temperature threshold configuration, where
    configuring this setting requires updating one or more registers whose bits are
    assigned roles as either sign, integer, decimal, or filler bits.

    Attributes:
        int_val (int): the integer value to be written to this setting

        dec_val (DecBits): the decimal value to be written to this setting, represented
                        as a binary value (uint)

        num_int_bits (int): the number of bits assigned to integer values for this setting

        num_dec_bits (int): the number of bits assigned to decimal values for this setting

        num_int_bits (int): the number of bits assigned to filler values for this setting

        start_addx (int): address of the MSB for this setting

        setting_name (TempType): the name for this temperature setting
    """

    int_val: int
    dec_val: int
    num_int_bits: int
    num_dec_bits: int
    num_filler_bits: int
    start_addx: int
    setting_name: TempType


@dataclass
class TempRange:
    """
    Thermocouple temperature range of allowed values by thermocouple type.

    Attributes:

        tc_low (int): hot-junction lowest temperature

        tc_high (int): hot-junction highest temperature

        cold_junct_low (int): cold-junction lowest temperature

        cold_junct_high (int): cold-junction highest temperature

    """

    tc_low: int
    tc_high: int
    cold_junct_low: int
    cold_junct_high: int


_tc_type_temps = {
    TCType.TYPE_B.value.op_code: TempRange(250, 1820, 0, 125),
    TCType.TYPE_E.value.op_code: TempRange(-200, 1000, -55, 125),
    TCType.TYPE_J.value.op_code: TempRange(-210, 1200, -55, 125),
    TCType.TYPE_K.value.op_code: TempRange(-200, 1372, -55, 125),
    TCType.TYPE_N.value.op_code: TempRange(-200, 1300, -55, 125),
    TCType.TYPE_R.value.op_code: TempRange(-50, 1768, -50, 125),
    TCType.TYPE_S.value.op_code: TempRange(-50, 1768, -50, 125),
    TCType.TYPE_T.value.op_code: TempRange(-200, 400, -55, 125),
}


def _format_temp_bitstring(
    sign_bit: int, int_val: int, dec_val: int, num_int_bits: int, num_dec_bits: int, fill_bits: int
):
    """Combine and place in order the sign, int, dec, and filler bits of a TempCode into a BitArray

    Returns:
        a bitstring BitArray ordered from MSB to LSB as a MAX31856 temperature register: sign_bit,
        int_val, dec_val, fill_bits.
    """
    bits = BitArray(uint=sign_bit, length=1)

    int_bits = BitArray(uint=int_val, length=num_int_bits)
    bits.append(int_bits)
    dec_bits = BitArray(uint=dec_val, length=num_dec_bits) if num_dec_bits else ""
    bits.append(dec_bits)

    # pylint: disable=unused-variable
    for i in range(fill_bits):
        bits += "0b0"
    return bits


def _slice_bitstring_to_opcodes(temp_code: TempCode, bitstr: BitArray, num_slices: int) -> list:
    """Uses a bitstring representing a temperature setting update to produce OpCodes
    for carrying out this operation

    Returns:
        a list of OpCodes
    """
    op_codes = []
    for i in range(num_slices):
        start = i * 8
        end = start + 8
        value = bitstr[start:end].uint
        # create new OpCode, incrementing address to next register each time if
        # this temp_code requires > 1 register.
        op_codes.append(OpCode(value, temp_code.start_addx + i, Masks.BYTE_MASK.value))
    return op_codes


class TempOutOfRangeError(ValueError):
    """Raised when a temperature threshold entered by the user is out of range for the currently
    configured thermocouple type.
    """


class IncompleteTempError(ValueError):
    """Raised when the user sets only either the integer or decimal value
    for a temperature threshold which requires both.
    """


class MissingTCTypeError(ValueError):
    """Raised if no thermocouple type is passed to tempcode_to_opcode"""


class IncompatibleRegisterSizeError(ValueError):
    """Raised when the number of bits contained in a TempCode is not a multiple
    of MAX31856 register sizes (8 bits).
    """


class IllegalTempTypeError(ValueError):
    """Raised when a non-existent TempType is entered"""


class ColdJunctionOverWriteError(Exception):
    """
    Raised when the user attempts to write temperature values to the cold-junction
    sensor without first having disabled it."""


def _dec_bits_to_float(dec_bits: int, num_dec_bits: int, is_negative: bool = False):
    """converts a decimal value formatted as DecBits4 or DecBits6 into float

    Args:
        dec_bits (int): the uint value representing the binary code for decimal value,
                        obtained from a DecBits4 or DecBits6 enum.

        num_dec_bits (int): the number of bits assigned for decimal values

        is_negative (bool): is this a negative decimal value or not
    """
    # shift the bits num_dec_bits places behind the decimal point
    if num_dec_bits == 0:
        return 0

    return (dec_bits * 2**-num_dec_bits) * (-1 if is_negative else 1)


def _validate_temperatures(tempcode: TempCode, tc_type: TCType):
    """Validates that the integer value of a TempCode is within the permitted range for
    the targeted register or thermocouple type.
    """
    # use thermocouple type to get TempRange object from dict
    temps = _tc_type_temps[tc_type.value.op_code]

    temp_ranges = {
        TempType.COLD_JUNCTION.value: {"min": temps.cold_junct_low, "max": temps.cold_junct_high},
        TempType.THERMOCOUPLE.value: {"min": temps.tc_low, "max": temps.tc_high},
        TempType.COLD_JUNCTION_OFFSET.value: {"min": -7, "max": 7},
    }
    is_negative = bool(tempcode.int_val < 0)
    temp_type = tempcode.setting_name.value
    temp_val = tempcode.int_val + _dec_bits_to_float(
        tempcode.dec_val.value, tempcode.num_dec_bits, is_negative
    )

    if temp_type not in temp_ranges:
        raise IllegalTempTypeError(f"TempType {temp_type} does not exist")

    if not temp_ranges[temp_type]["min"] <= temp_val <= temp_ranges[temp_type]["max"]:
        raise TempOutOfRangeError(
            f"""Temperature integer value {temp_val} exceeds writeable limits
        for setting {tempcode.setting_name} for {tc_type} thermocouple"""
        )


def tempcode_to_opcode(temp_code: TempCode, tc_type: TCType, cj_status: bool) -> list:
    """
    Generates a list of OpCodes necessary to execture the temperature setting
    updates contained in a TempCode.

    Returns:
        A list of OpCode objects

    Raises:
        ValueError: if temp_code is None

        IncompleteTempError: if TempCode object only has either an integer or decimal value

        MissingTCTypeError: if thermocouple type is not provided

        IncompatibleRegisterSizeError: if number of bits in TempCode object is not divisible by
                                        MAX31856 register size of 8 bits.

        ColdJunctionOverWriteError: if value is written to cold-junction temperature registers
                                    while cold-junction sensing is not disabled.
    """
    # pylint: disable=too-many-branches

    if temp_code is None:
        raise ValueError("temp_code must be of type TempCode: received None")

    # no args passed for this temp setting -- skip it
    # second clause is to check for registers with no decimal bits,
    # since dec_val will never be None
    if temp_code.int_val is None and (temp_code.dec_val is None or temp_code.num_dec_bits == 0):
        return []

    # only either int or dec args passed for this temp setting
    if temp_code.int_val is None or temp_code.dec_val is None:
        raise IncompleteTempError(
            f"temp_code {temp_code} requires both int and dec values: received only one of these."
        )

    if tc_type is None:
        raise MissingTCTypeError(
            "thermocouple type is required to compute thermocouple temperature range"
        )

    if not cj_status and temp_code.start_addx == TCAddresses.CJTH_W.value:
        raise ColdJunctionOverWriteError(
            "Cold-junction sensor must be disabled in order to write temperature values to it"
        )

    # validate temp range
    _validate_temperatures(temp_code, tc_type)

    # compute total bits required by this setting
    num_bits = 1 + temp_code.num_int_bits + temp_code.num_dec_bits + temp_code.num_filler_bits

    if num_bits % REG_SIZE:
        raise IncompatibleRegisterSizeError(
            "Number of temp_code bits not divisible by 8 bit register size"
        )

    # check if negative temperature
    sign_bit = 0
    if temp_code.int_val < 0:
        sign_bit = 1
        temp_code.int_val = abs(temp_code.int_val)

    # combine and place in order the sign, int, dec, and filler bits into a BitArray
    bits = _format_temp_bitstring(
        sign_bit,
        temp_code.int_val,
        temp_code.dec_val.value,
        temp_code.num_int_bits,
        temp_code.num_dec_bits,
        temp_code.num_filler_bits,
    )

    # slice bitstring into 8-bit register sized chunks and convert each chunk to OpCode
    num_regs = int(num_bits / REG_SIZE)  # number of registers this setting uses
    return _slice_bitstring_to_opcodes(temp_code, bits, num_regs)



"""
Helper functions and class for modifying device register values

Classes:

    OpCode
    OpCodeMaskIncompatibleError(ValueError)

Functions:

    _add_change_flags(dict)
    apply_opcodes(dict, list)
    apply_opcode(OpCode, int)
"""


from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
import logging


_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class OpCode:
    """
    Represents an operation to opcode to the value stored in a memory register

    Attributes
    ----------
    `op_code` : int
        the value used to update the register bits relevant to this setting or command.
        Please see below for a practical example on how to define the op_code.
    `reg_address` : int
        the register's address in the device's register memory map
    `op_mask` : int
        the value used to clear the register bits whose value will be updated
        when the op_code is applied. Mask bits intended to clear corresponding
        bits in the target value should be assigned a value of 0.
        Please see below for a practical example on how to define the op_mask.

    An example of how to define OpCode fields:

        Assume we have a register with the following bit values: register_value = (0b1010 1101)

        Let's assume we wish to update only the high bits (0b1010) to a new value, (0b0110).
        So our intended goal is to change the register_value to (0b0110 1101).

        To do so, we first need to clear, or in other words, mask these high bits.
        The mask to use here should be defined as (0b0000 1111).
        This allows us to clear the high bits in a single AND operation, as follows:

            (0b1010 1101) & (0b0000 1111) = (0b0000 1101).

        Now the high bits have been cleared, we can apply our opcode. In order to obtain our final
        register_value of (0b0110 1101), our opcode should be (0b0110 0000). Applying this opcode
        to our now 'masked' register_value, with an OR operation, gives us our final register value:

            (0b0000 1101) | (0b0110 0000) = (0b0110 1101)
    """

    op_code: int
    reg_address: int
    op_mask: int


class OpCodeMaskIncompatibleError(ValueError):
    """Raised when an OpCode contains an op_code which affects bits not covered by the op_mask"""


class RegisterUpdateError(Exception):
    """
    Raised if the value of an unmodified register has been changed
    while opcodes are being applied.
    """


def apply_opcodes(register_values: dict, opcodes: list):
    """
    Generates updated register values after applying opcodes,
    and sets is_changed flag for updated registers.

    Args:
        register_values (dict): a map of a device's registers to a dictionary containing their
                                corresponding register values. The dictionary must contain
                                entries of the form register_address (int): register_value (int).

        opcodes (list): a list of OpCode objects to be used for updating register values.

    Returns:
        `dict`: a map of the device's registers to a dictionary containing
        the updated values and change flags, formatted as:

            { reg_addx (int): {"value": int, "is_changed": bool} }

    Raises:
        ValueError: if either register_values or opcodes is empty
    """
    if len(register_values) < 1 or len(opcodes) < 1:
        _logger.error(
            "empty values received for 'register_values' or 'opcodes' args, opcodes not applied"
        )
        raise ValueError("register_values and opcodes args must both be non-empty")
    _format_register_map(register_values)

    original_regs = deepcopy(register_values)

    # apply each opcode to its corresponding register
    for opcode in opcodes:
        register_entry = register_values.get(opcode.reg_address)
        # if this opcode maps to a valid register address
        if register_entry is not None:
            # apply the opcode to the register
            register_entry["value"] = _apply_opcode(register_entry["value"], opcode)
            register_entry["is_changed"] = True

    __validate_register_updates(original_regs, register_values)

    return register_values


def _apply_opcode(register_value: int, opcode: OpCode):
    """
    Generates an update code for a specific register by applying an opcode

    Args:
        opcode (OpCode): an OpCode object representing the opcode to be applied to the register

        register_value (int): the current stored value of the register

    Returns:
        an update code, the value to be written back to the register in order to apply
        the opcode.
    """
    # ensure op_code only writes to bits of register covered by mask
    if (~opcode.op_mask & opcode.op_code) != opcode.op_code:
        raise OpCodeMaskIncompatibleError(
            f"""opcode ({hex(opcode.op_code)}) affects bits
            not covered by mask ({hex(opcode.op_mask)}"""
        )

    register_value &= opcode.op_mask  # clear the bits to be overwritten
    register_value |= opcode.op_code  # apply the opcode to the cleared bits

    return register_value


def __validate_register_updates(original_regs, updated_regs):
    """
    Verifies the value of regsiters not targeted for updates has not changed after applying updates
    """
    for addx, entry in updated_regs.items():
        if not entry["is_changed"] and entry["value"] != original_regs[addx]["value"]:
            raise RegisterUpdateError(
                f"Register at address {addx} has been incorrectly targeted for updates."
            )


def _add_change_flags(register_values: dict):
    """
    adds flags to register values for checking to see if register value has been modified later

    Args:
        register_values (dict): a map of the device's registers to their current values

    Returns:
        the same dictionary but with an added 'is_changed' field for each entry
    """
    for key in register_values:
        register_values[key]["is_changed"] = False


def _convert_values_to_dict(reg_map: dict) -> dict:
    """Changes a register_address, register_value map to use register_address, dictionary format
    where the dictionary now hold the register_value.

    Args:
        reg_map (dict): a dictionary containing (int) register_address, (int) register_value pairs

    Returns:
        a dictionary containing (int) register_address, {'value': register_value} pairs
    """
    # convert each addx, value pair to addx, dictionary pair
    for addx, value in reg_map.items():
        reg_map[addx] = {"value": value}


def _format_register_map(reg_map: dict) -> dict:
    """converts map of register_address, register_value pairs to a format compatible with functions
    defined in this module.

    Args:
        reg_map (dict): a dictionary containing (int) register_address, (int) register_value pairs

    Returns:
        a dictionary holding (int) reg_address, {'value': reg_value, 'is_changed': False} pairs
    """
    _convert_values_to_dict(reg_map)
    _add_change_flags(reg_map)


def convert_dict_to_values(reg_dict: dict = None):
    """
    Function to re-formate register dictionary back to original form
    In:
        reg_dict (dict): register address to value and is_changed flag
                        {register_address : {'value' : value(int), is_changed : bool}}
        Returns:
            reg_dict (dict): register address to value
                             {register_address : value}
    """
    for reg_addx, entry in reg_dict.items():
        reg_dict[reg_addx] = entry["value"]
    return reg_dict


def is_bit_set(reg_val: int, bit_mask: int):
    """
    Get state of a bit from an 8-bit register

    Args:
        reg_val (int): 8-bit value of the register this bit belongs to
        bit_mask (int): value used to mask all other bits in this register.
            Hint: use bit mask corresponding to opcode used to set this bit.

    Returns:
        `bool`: True if the bit is set, False otherwise.
    """
    return bool(reg_val & (~bit_mask))


class BitMask(Enum):
    """Bit/Byte masks for use with OpCodes"""

    LOW_NIBBLE = 0xF0  # overwrite low nibble bits, let high nibble bits pass
    HIGH_NIBBLE = 0x0F  # overwrite high nibble bits, let low nibble bits pass
    BYTE = 0x00 # overwrite all bits
    BIT0 = 0xFE # overwrite bit 0
    BIT1 = 0xFD # overwrite bit 1
    BIT2 = 0xFB # overwrite bit 2
    BIT3 = 0xF7 # overwrite bit 3
    BIT4 = 0xEF # overwrite bit 4
    BIT5 = 0xDF # overwrite bit 5
    BIT6 = 0xBF # overwrite bit 6
    BIT7 = 0x7F # overwrite bit 7



'''
Provides a class for interacting with the GPIO pins through GPIO peripheral
'''

import logging
from edgepi.peripherals.gpio import GpioDevice
from edgepi.gpio.gpio_constants import GpioDevPaths
from edgepi.gpio.gpio_configs import DOUTPins, DINPins, generate_gpiochip_pin_info

_logger = logging.getLogger(__name__)

class EdgePiGPIOChip(GpioDevice):
    """
    A class to represent the GPIO peripheral using gpiochip device. This class will be imported to
    each module that requires GPIO manipulation.
    """
    # dictionary mapping pin name to CPU gpio pin number
    __pin_name_dict = {DINPins.DIN1.value : 26,
                       DINPins.DIN2.value : 6,
                       DINPins.DIN3.value : 11,
                       DINPins.DIN4.value : 9,
                       DINPins.DIN5.value : 22,
                       DINPins.DIN6.value : 27,
                       DINPins.DIN7.value : 3,
                       DINPins.DIN8.value : 2,
                       DOUTPins.DOUT1.value : 13,
                       DOUTPins.DOUT2.value : 12}

    def __init__(self):
        super().__init__(GpioDevPaths.GPIO_CIHP_DEV_PATH.value)
        self.gpiochip_pins_dict = generate_gpiochip_pin_info()

    def read_gpio_pin_state(self, pin_name: str = None):
        """
        Read current state of GPIO pins. If the GPIO object is instantiated, it will be a unique
        object until close() method is called. So every time the state is read, it will instantiate
        before read and cloase() after read.
        Args:
            pin_name (str): name of the pin
        Returns:
            `bool`: True if state is high, False if state is low
        """
        self.open_gpio(pin_num=self.__pin_name_dict[pin_name],
                       pin_dir=self.gpiochip_pins_dict[pin_name].dir,
                       pin_bias=self.gpiochip_pins_dict[pin_name].bias)
        state = self.read_state()
        self.close()
        return state

    def write_gpio_pin_state(self, pin_name: str = None, state: bool = None):
        """
        write pin state
        Args:
            pin_name (str): name of the pin to write state to
            state (bool): state to write, True = High, False = Low
        Return:
            N/A
        """
        self.open_gpio(pin_num=self.__pin_name_dict[pin_name],
                       pin_dir=self.gpiochip_pins_dict[pin_name].dir,
                       pin_bias=self.gpiochip_pins_dict[pin_name].bias)
        self.write_state(state)
        read_back = self.read_state()
        self.close()
        return read_back

    def set_gpio_pin_dir(self, pin_name: str = None, direction: bool = None):
        """
        Set gpio pin direction
        Args:
            pin_name (str): name of the pin
            direction (bool): direction to write, True = Input, False = Output
        """
        self.open_gpio(pin_num=self.__pin_name_dict[pin_name],
                       pin_dir="in" if direction else "out",
                       pin_bias=self.gpiochip_pins_dict[pin_name].bias)
        self.close()

    def toggle_gpio_pin_state(self, pin_name: str = None):
        """
        Toggle pin state
        Args:
            pin_name (str): name of the pin to write state to
        Return:
            N/A
        """
        self.open_gpio(pin_num=self.__pin_name_dict[pin_name],
                       pin_dir=self.gpiochip_pins_dict[pin_name].dir,
                       pin_bias=self.gpiochip_pins_dict[pin_name].bias)
        state = self.read_state()
        if state:
            self.write_state(False)
        else:
            self.write_state(True)
        self.close()



'''Helper class to access on board eeprom'''

import logging
import math

from edgepi.calibration.eeprom_constants import (
    EEPROMInfo,
    EdgePiMemoryInfo,
    MessageFieldNumber,
    EdgePiEEPROMData)
from edgepi.calibration.eeprom_mapping_pb2 import EepromLayout
from edgepi.peripherals.i2c import I2CDevice


class EdgePiEEPROM(I2CDevice):
    '''
    Helper class to read eeprom using I2C
    '''
    __dev_path = '/dev/i2c-10'

    def __init__(self):
        self.log = logging.getLogger(__name__)
        self.eeprom_layout = EepromLayout()
        super().__init__(self.__dev_path)

    def __pack_mem_address(self, page_addr: int = None, byte_addr: int = None):
        """
        Pack page address and byte address to generate address message. The address message will
        let the EEPROM memory address point to know at which address the read/write operation to
        take in place
        Args:
            page_addr (int): page address (0~511)
            byte_addr (int): byte address in the page(0~63)
        return:
            (list): 2-byte address message
        """
        address = page_addr<<6 | byte_addr
        return [(address>>8)&0xFF, address&0xFF]

    def __byte_address_generation(self, memory_address: int = None):
        """
        Generates page address and bytes address from the memory address provided.
        Args:
            memory_address (int): memory address to start read/write operation. This is in bytes
            from 0 - 32768
        """
        page_addr  = math.floor(memory_address/EEPROMInfo.PAGE_SIZE.value)
        byte_addr = memory_address%EEPROMInfo.PAGE_SIZE.value
        self.log.debug(f'Page address = {page_addr}, byte Address = {byte_addr}')
        return page_addr, byte_addr

    def __allocated_memory(self):
        '''
        The first two bytes represenst the allocated memory in Edgepi reserved memory space. This
        function returns the length of memory to read.
        Args:
            N/A
        Return:
            length (int): size of memory to read
        '''
        length = self.sequential_read(EdgePiMemoryInfo.USED_SPACE.value, 2)
        return (length[0]<<8)| length[1]

    def __read_edgepi_reserved_memory(self):
        '''
        Read Edgepi reserved memory space to retreive parameters. This function will return byte
        strings, that can be converted into protocol buffer message format
        Args:
            N/A
        Return:
            Byte_string (bytes): strings of bytes read from the eeprom
        '''
        mem_size = self.__allocated_memory()
        buff_list = self.sequential_read(EdgePiMemoryInfo.BUFF_START.value, mem_size)
        return bytes(buff_list)

    def get_message_of_interest(self, msg: MessageFieldNumber = None):
        """
        This function filters out the message according to the specified field number passed as
        parameter.
        Args:
            msg (MessageFieldNumber): protocol buffer message field index number for ListFields()
            function
        Return:
            pb message specified by the message field number. ex) if message field of DAC is passed,
            the dac message will be returned
        """
        self.eeprom_layout.ParseFromString(self.__read_edgepi_reserved_memory())
        return self.eeprom_layout.ListFields()[msg.value - 1][1]

    def get_edgepi_reserved_data(self):
        """
        Read Edgepi reserved memory space and populate dataclass
        Args:
            N/A
        Return:
            eeprom_data (EdgePiEEPROMData): dataclass containing eeprom values
        """
        # pylint: disable=no-member
        self.eeprom_layout.ParseFromString(self.__read_edgepi_reserved_memory())
        eeprom_data = EdgePiEEPROMData()
        eeprom_data.dac_calib_parms=eeprom_data.message_to_dict(self.eeprom_layout.dac)
        eeprom_data.adc_calib_parms=eeprom_data.message_to_dict(self.eeprom_layout.adc)
        eeprom_data.rtd_calib_parms=eeprom_data.message_to_dict(self.eeprom_layout.rtd)
        eeprom_data.tc_calib_parms=eeprom_data.message_to_dict(self.eeprom_layout.tc)
        eeprom_data.config_key=eeprom_data.keys_to_str(self.eeprom_layout.config_key)
        eeprom_data.data_key=eeprom_data.keys_to_str(self.eeprom_layout.data_key)
        eeprom_data.serial= self.eeprom_layout.serial_number
        eeprom_data.model= self.eeprom_layout.model
        eeprom_data.client_id= self.eeprom_layout.client_id
        return eeprom_data

    def sequential_read(self, mem_addr: int = None, length: int = None):
        '''
        Read operation reads the specified number of memory location starting from provided address.
        The address pointer will wrap around when it reaches the end of the memory.
        Args:
            mem_addr: starting memory address to read from
            len: length of data or number of data to read
        Returns:
            List of read data

        '''
        page_addr, byte_addr = self.__byte_address_generation(mem_addr)
        mem_addr_list = self.__pack_mem_address(page_addr, byte_addr)
        msg = self.set_read_msg(mem_addr_list, [0x00]*length)
        self.log.debug(f'Reading Address {mem_addr}, {length} bytes, {msg[1].data}')
        read_result = self.transfer(EEPROMInfo.DEV_ADDR.value, msg)
        self.log.debug(f'Read data: {msg[1].data}')
        return read_result


    def selective_read(self, mem_addr: int = None):
        '''
        Read operation reads a data from the specified address
        Args:
            mem_addr: starting memory address to read from
        Returns:
            List of read data
        '''
        page_addr, byte_addr = self.__byte_address_generation(mem_addr)
        mem_addr_list = self.__pack_mem_address(page_addr, byte_addr)
        msg = self.set_read_msg(mem_addr, [0x00])
        self.log.debug(f'Reading Address {mem_addr_list}, {msg[1].data}')
        read_result = self.transfer(EEPROMInfo.DEV_ADDR.value, msg)
        self.log.debug(f'Read data: {msg[1].data}')
        return read_result

    def byte_write_register(self, mem_addr: int = None, data: int = None):
        '''
        Write operation writes a data to the specified address
        Args:
            mem_addr: starting memory address to read from
            data: data to write to the location
        Returns:
            N/A
        '''
        page_addr, byte_addr = self.__byte_address_generation(mem_addr)
        mem_addr_list = self.__pack_mem_address(page_addr, byte_addr)
        msg = self.set_write_msg(mem_addr_list, [data])
        self.log.debug(f'Writing {data} to memory address of {mem_addr}, {msg[0].data}')
        self.transfer(EEPROMInfo.DEV_ADDR.value, msg)

    def page_write_register(self, mem_addr: int = None, data: list = None):
        '''
        Write operation writes a page of data to the specified address
        Args:
            mem_addr: starting memory address to read from
            data: data to write to the location
        Returns:
            N/A
        '''
        page_addr, byte_addr = self.__byte_address_generation(mem_addr)
        mem_addr_list = self.__pack_mem_address(page_addr, byte_addr)
        msg = self.set_write_msg(mem_addr_list, data)
        self.log.debug(f'Writing {data} to memory address of {mem_addr}, {msg[0].data}')
        self.transfer(EEPROMInfo.DEV_ADDR.value, msg)



"""unit tests for gpio.py module"""
# pylint: disable=wrong-import-position
from unittest import mock
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()
import pytest
from edgepi.peripherals.gpio import GpioDevice


@pytest.mark.parametrize(
    "fd",
    [
        ("/dev/gpiochip0", 27, "in", "pull_down"),
    ],
)
def test_gpio_init_param(mocker, fd):
    mocker.patch("edgepi.peripherals.gpio.GPIO")
    gpio = GpioDevice(fd)
    assert gpio.fd == fd
    assert gpio.gpio is None

@pytest.mark.parametrize(
    "fd, result",
    [
        ("/dev/gpiochip0", True),
        ("/dev/gpiochip0", False),
        ("/dev/gpiochip0",True),
        ("/dev/gpiochip0", False),
    ],
)
def test_gpio_read_state(mocker, fd, result):
    mocker.patch("edgepi.peripherals.gpio.GpioDevice.read_state", return_value = result)
    gpio = GpioDevice(fd)
    assert gpio.read_state() == result



'''unit test for access eeprom'''
# pylint: disable=C0413
# pylint: disable=no-member
from unittest import mock
import os
PATH = os.path.dirname(os.path.abspath(__file__))
import sys
sys.modules['periphery'] = mock.MagicMock()

from edgepi.calibration.eeprom_constants import EdgePiEEPROMData, Keys
from edgepi.calibration.eeprom_mapping_pb2 import EepromLayout

def read_binfile():
    """Read the dummy serializedFile and return byte string"""
    with open(PATH+"/serializedFile","rb") as fd:
        b_string = fd.read()
    return b_string

def test_edgepi_eeprom_data():
    memory_map = EepromLayout()
    memory_map.ParseFromString(read_binfile())
    eeprom_data = EdgePiEEPROMData()
    eeprom_data.dac_calib_parms=eeprom_data.message_to_dict(memory_map.dac)
    eeprom_data.adc_calib_parms=eeprom_data.message_to_dict(memory_map.adc)
    eeprom_data.rtd_calib_parms=eeprom_data.message_to_dict(memory_map.rtd)
    eeprom_data.tc_calib_parms=eeprom_data.message_to_dict(memory_map.tc)
    eeprom_data.config_key=eeprom_data.keys_to_str(memory_map.config_key)
    eeprom_data.data_key=eeprom_data.keys_to_str(memory_map.data_key)
    eeprom_data.serial=memory_map.serial_number
    eeprom_data.model=memory_map.model
    eeprom_data.client_id=memory_map.client_id
    assert isinstance(eeprom_data.dac_calib_parms, dict)
    assert isinstance(eeprom_data.adc_calib_parms, dict)
    assert isinstance(eeprom_data.rtd_calib_parms, dict)
    assert isinstance(eeprom_data.tc_calib_parms, dict)
    assert isinstance(eeprom_data.config_key, Keys)
    assert isinstance(eeprom_data.config_key, Keys)
    assert isinstance(eeprom_data.data_key, Keys)
    assert isinstance(eeprom_data.data_key, Keys)
    assert isinstance(eeprom_data.serial, str)
    assert isinstance(eeprom_data.model, str)
    assert isinstance(eeprom_data.client_id, str)
    assert eeprom_data.dac_calib_parms == eeprom_data.message_to_dict(memory_map.dac)
    assert eeprom_data.adc_calib_parms==eeprom_data.message_to_dict(memory_map.adc)
    assert eeprom_data.rtd_calib_parms==eeprom_data.message_to_dict(memory_map.rtd)
    assert eeprom_data.tc_calib_parms==eeprom_data.message_to_dict(memory_map.tc)
    assert eeprom_data.config_key.certificate == memory_map.config_key.certificate
    assert eeprom_data.config_key.private == memory_map.config_key.private_key
    assert eeprom_data.data_key.certificate == memory_map.data_key.certificate
    assert eeprom_data.data_key.private == memory_map.data_key.private_key
    assert eeprom_data.serial == memory_map.serial_number
    assert eeprom_data.model == memory_map.model
    assert eeprom_data.client_id == memory_map.client_id



'''
Utility module for computing MAX31856 conversion time

Functions:
    calc_conv_time(int, int, bool)
'''


from enum import Enum
from bitstring import Bits

SAFETY_MARGIN = 1.1


class ConvTimes(Enum):
    """Converson time delays for various MAX31856 settings configurations"""

    CJ_ON_OC_LOW_NOM = 13.3
    CJ_ON_OC_LOW_MAX = 15
    CJ_ON_OC_MED_NOM = 33.4
    CJ_ON_OC_MED_MAX = 37
    CJ_ON_OC_HIGH_NOM = 113.4
    CJ_ON_OC_HIGH_MAX = 125
    CJ_OFF_OC_LOW_NOM = 40
    CJ_OFF_OC_LOW_MAX = 44
    CJ_OFF_OC_MED_NOM = 60
    CJ_OFF_OC_MED_MAX = 66
    CJ_OFF_OC_HIGH_NOM = 140
    CJ_OFF_OC_HIGH_MAX = 154
    CJ_OFF = -25.0
    DEFAULT_HZ50_NOM = 169
    DEFAULT_HZ50_MAX = 185
    DEFAULT_HZ60_NOM = 143
    DEFAULT_HZ60_MAX = 155
    HZ50_AVG_2 = 40.33
    HZ50_AVG_4 = 120.99
    HZ50_AVG_8 = 282.31
    HZ50_AVG_16 = 604.95
    HZ60_AVG_2 = 33.33
    HZ60_AVG_4 = 99.99
    HZ60_AVG_8 = 233.31
    HZ60_AVG_16 = 499.95


_open_circ_delays = {
    0b100: (0, 0),
    0b101: (ConvTimes.CJ_OFF_OC_LOW_NOM.value, ConvTimes.CJ_OFF_OC_LOW_MAX.value),
    0b110: (ConvTimes.CJ_OFF_OC_MED_NOM.value, ConvTimes.CJ_OFF_OC_MED_MAX.value),
    0b111: (ConvTimes.CJ_OFF_OC_HIGH_NOM.value, ConvTimes.CJ_OFF_OC_HIGH_MAX.value),
    0b000: (0, 0),
    0b001: (ConvTimes.CJ_ON_OC_LOW_NOM.value, ConvTimes.CJ_ON_OC_LOW_MAX.value),
    0b010: (ConvTimes.CJ_ON_OC_MED_NOM.value, ConvTimes.CJ_ON_OC_MED_MAX.value),
    0b011: (ConvTimes.CJ_ON_OC_HIGH_NOM.value, ConvTimes.CJ_ON_OC_HIGH_MAX.value),
}

_default_times = {
    0b0: (ConvTimes.DEFAULT_HZ60_NOM.value, ConvTimes.DEFAULT_HZ60_MAX.value),
    0b1: (ConvTimes.DEFAULT_HZ50_NOM.value, ConvTimes.DEFAULT_HZ50_MAX.value),
}

_avg_mode_delays = {
    0b00000: 0,
    0b10000: 0,
    0b00001: ConvTimes.HZ60_AVG_2.value,
    0b00010: ConvTimes.HZ60_AVG_4.value,
    0b00011: ConvTimes.HZ60_AVG_8.value,
    0b00100: ConvTimes.HZ60_AVG_16.value,
    0b10001: ConvTimes.HZ50_AVG_2.value,
    0b10010: ConvTimes.HZ50_AVG_4.value,
    0b10011: ConvTimes.HZ50_AVG_8.value,
    0b10100: ConvTimes.HZ50_AVG_16.value,
}


def calc_conv_time(cr0_val: int, cr1_val: int, safe_delay: bool = True):
    """Computes the time delay required for thermocouple single sampling

    Args:
        cr0_val (int): the value stored in register CR0

        cr1_val (int): the value stored in register CR1

        safe_delay (bool): indicator for whether to compute min or max time delay

    Returns:
        a float representing the computed time delay in milliseconds (ms)
    """
    cr0 = Bits(uint=cr0_val, length=8)
    cr1 = Bits(uint=cr1_val, length=8)
    noise_filt_bit = cr0[7]
    cold_junct_bit = cr0[4]
    open_circ_bits = cr0[2:4]
    avg_mode_bits = cr1[0:4]

    conv_time = 0

    # format bit strings for various modes
    open_circ_bit_str = [cold_junct_bit] + open_circ_bits
    avg_mode_bit_str = [noise_filt_bit] + avg_mode_bits

    # compute conv time for settings that depend on safety margin
    if safe_delay:
        conv_time += _default_times[noise_filt_bit][1]  # base conv time
        conv_time += _open_circ_delays[open_circ_bit_str.uint][1]  # open circuit test delay
    else:
        conv_time += _default_times[noise_filt_bit][0]  # base conv time
        conv_time += _open_circ_delays[open_circ_bit_str.uint][0]  # open circuit test delay

    if cold_junct_bit:
        conv_time += ConvTimes.CJ_OFF.value

    # additional conv time delays
    conv_time += _avg_mode_delays[avg_mode_bit_str.uint]

    return conv_time * SAFETY_MARGIN
